@@ -27,6 +27,7 @@
 
 #include "pkcs11/pkcs11.h"
 #include "pkcs11/pkcs11g.h"
+#include "pkcs11/pkcs11i.h"
 
 #include <glib.h>
 
@@ -60,10 +61,11 @@ auth_object_for_context_specific (CK_SESSION_HANDLE handle, CK_OBJECT_HANDLE obj
 {
 	GkrPkcs11AuthObject *info = NULL;
 	CK_SESSION_INFO session_info;
-	CK_ATTRIBUTE attrs[5];
+	CK_ATTRIBUTE attrs[6];
 	CK_OBJECT_CLASS klass;
 	gchar *label = NULL;
 	gchar *unique = NULL;
+	gchar *digest = NULL;
 	CK_BBOOL token, always;
 	CK_ULONG n_attrs;
 	CK_RV rv;
@@ -77,21 +79,26 @@ auth_object_for_context_specific (CK_SESSION_HANDLE handle, CK_OBJECT_HANDLE obj
 	attrs[1].pValue = unique = NULL;
 	attrs[1].ulValueLen = 0;
 
-	attrs[2].type = CKA_CLASS;
-	attrs[2].pValue = &klass;
-	attrs[2].ulValueLen = sizeof (klass);
+	/* COMPAT: Loaded for compatibility with old gnome-keyrings */
+	attrs[2].type = CKA_GNOME_INTERNAL_SHA1;
+	attrs[2].pValue = digest = NULL;
+	attrs[2].ulValueLen = 0;
+
+	attrs[3].type = CKA_CLASS;
+	attrs[3].pValue = &klass;
+	attrs[3].ulValueLen = sizeof (klass);
 
 	always = CK_FALSE;
-	attrs[3].type = CKA_ALWAYS_AUTHENTICATE;
-	attrs[3].pValue = &always;
-	attrs[3].ulValueLen = sizeof (always);
+	attrs[4].type = CKA_ALWAYS_AUTHENTICATE;
+	attrs[4].pValue = &always;
+	attrs[4].ulValueLen = sizeof (always);
 	
 	token = CK_FALSE;
-	attrs[4].type = CKA_TOKEN;
-	attrs[4].pValue = &token;
-	attrs[4].ulValueLen = sizeof (token);
+	attrs[5].type = CKA_TOKEN;
+	attrs[5].pValue = &token;
+	attrs[5].ulValueLen = sizeof (token);
 	
-	n_attrs = 5;
+	n_attrs = 6;
 	
 	/* Get attribute sizes */
 	rv = (pkcs11_lower->C_GetAttributeValue) (handle, object, attrs, n_attrs);
@@ -112,12 +119,15 @@ auth_object_for_context_specific (CK_SESSION_HANDLE handle, CK_OBJECT_HANDLE obj
 		attrs[0].pValue = label = g_malloc0 (attrs[0].ulValueLen + 1);
 	if (attrs[1].ulValueLen != (CK_ULONG)-1)
 		attrs[1].pValue = unique = g_malloc0 (attrs[1].ulValueLen + 1);
+	if (attrs[2].ulValueLen != (CK_ULONG)-1)
+		attrs[2].pValue = digest = g_malloc0 (attrs[2].ulValueLen + 1);
 	
 	/* Get actual attributes */
 	rv = (pkcs11_lower->C_GetAttributeValue) (handle, object, attrs, n_attrs);
 	if (rv != CKR_OK && rv != CKR_ATTRIBUTE_TYPE_INVALID) {
 		g_free (label);
 		g_free (unique);
+		g_free (digest);
 		return NULL;
 	}
 	
@@ -132,6 +142,11 @@ auth_object_for_context_specific (CK_SESSION_HANDLE handle, CK_OBJECT_HANDLE obj
 		info->unique = unique;
 		unique = NULL;
 	}
+	
+	if (attrs[2].ulValueLen != (CK_ULONG)-1) {
+		info->digest = digest;
+		digest = NULL;
+	}
 
 	info->token = token;
 	info->klass = klass;
@@ -140,6 +155,7 @@ auth_object_for_context_specific (CK_SESSION_HANDLE handle, CK_OBJECT_HANDLE obj
 	
 	g_free (label);
 	g_free (unique);
+	g_free (digest);
 	
 	return info;
 }
@@ -209,52 +225,6 @@ auth_object_for_cache (CK_SESSION_HANDLE handle, CK_OBJECT_HANDLE object)
 	return info;
 }
 
-static CK_RV 
-perform_set_user_pin (CK_SESSION_HANDLE handle, CK_UTF8CHAR_PTR old_pin, CK_ULONG n_old_pin,
-                      CK_UTF8CHAR_PTR new_pin, CK_ULONG n_new_pin, gboolean also_login)
-{
-	CK_SESSION_INFO session_info;
-	CK_TOKEN_INFO token_info;
-	gboolean init_auth = FALSE;
-	CK_RV rv, login_rv;
-	
-	/* Dig up the information we'll need, and don't prompt if protected auth path */
-	if ((pkcs11_lower->C_GetSessionInfo) (handle, &session_info) == CKR_OK &&
-	    (pkcs11_lower->C_GetTokenInfo) (session_info.slotID, &token_info) == CKR_OK && 
-	    !(token_info.flags & CKF_PROTECTED_AUTHENTICATION_PATH)) { 
-
-		DAEMON_ENTER ();
-		
-			if (!(token_info.flags & CKF_USER_PIN_INITIALIZED))
-				init_auth = gkr_pkcs11_auth_init_user_prompt (handle, &token_info, &new_pin, &n_new_pin);
-			/* TODO: Prompt for other 'change password' case */
-
-		DAEMON_LEAVE ();
-	}
-
-	rv = (pkcs11_lower->C_SetPIN) (handle, old_pin, n_old_pin, new_pin, n_new_pin);
-	
-	/* If requested we can also login, this prevents two prompts */
-	login_rv = CKR_OK;
-	if (rv == CKR_OK) {
-		login_rv = (pkcs11_lower->C_Login) (handle, CKU_USER, new_pin, n_new_pin);
-	}
-	
-	if (init_auth) {
-		DAEMON_ENTER ();
-		
-			gkr_pkcs11_auth_init_user_done (handle, &token_info, &new_pin, &n_new_pin, rv);
-			/* TODO: Done for other case */
-			
-		DAEMON_LEAVE ();
-	}
-	
-	if (login_rv != CKR_OK)
-		rv = login_rv;
-	
-	return rv;
-}
-
 /* --------------------------------------------------------------------------------------
  * PKCS#11 ENTRY POINTS
  */
@@ -465,22 +435,73 @@ auth_C_GetSessionInfo (CK_SESSION_HANDLE handle, CK_SESSION_INFO_PTR info)
 }
 
 static CK_RV
-auth_C_InitPIN (CK_SESSION_HANDLE handle, CK_UTF8CHAR_PTR pin, CK_ULONG pin_len)
+auth_C_InitPIN (CK_SESSION_HANDLE handle, CK_UTF8CHAR_PTR pin, CK_ULONG n_pin)
 {
-	/* 
-	 * TODO: Need to implement this properly.
-	 * 
-	 * Prompt the user for new password if 
-	 * CKF_PROTECTED_AUTHENTICATION path.
-	 */
-	return (pkcs11_lower->C_InitPIN) (handle, pin, pin_len);
+	CK_SESSION_INFO session_info;
+	CK_TOKEN_INFO token_info;
+	gboolean init_auth = FALSE;
+	CK_RV rv;
+	
+	/* Dig up the information we'll need, and don't prompt if protected auth path */
+	if ((pkcs11_lower->C_GetSessionInfo) (handle, &session_info) == CKR_OK &&
+	    (pkcs11_lower->C_GetTokenInfo) (session_info.slotID, &token_info) == CKR_OK && 
+	    !(token_info.flags & CKF_PROTECTED_AUTHENTICATION_PATH)) { 
+
+		DAEMON_ENTER ();
+		
+			init_auth = gkr_pkcs11_auth_init_user_prompt (handle, &token_info, &pin, &n_pin);
+
+		DAEMON_LEAVE ();
+	}
+
+	rv = (pkcs11_lower->C_InitPIN) (handle, pin, n_pin);
+	
+	if (init_auth) {
+		DAEMON_ENTER ();
+		
+			gkr_pkcs11_auth_init_user_done (handle, &token_info, &pin, &n_pin, rv);
+			
+		DAEMON_LEAVE ();
+	}
+	
+	return rv;
 }
 
 static CK_RV
 auth_C_SetPIN (CK_SESSION_HANDLE handle, CK_UTF8CHAR_PTR old_pin, CK_ULONG n_old_pin, 
                CK_UTF8CHAR_PTR new_pin, CK_ULONG n_new_pin)
 {
-	return perform_set_user_pin (handle, old_pin, n_old_pin, new_pin, n_new_pin, FALSE);
+	CK_SESSION_INFO session_info;
+	CK_TOKEN_INFO token_info;
+	gboolean init_auth = FALSE;
+	CK_RV rv;
+	
+	/* Dig up the information we'll need, and don't prompt if protected auth path */
+	if ((pkcs11_lower->C_GetSessionInfo) (handle, &session_info) == CKR_OK &&
+	    (pkcs11_lower->C_GetTokenInfo) (session_info.slotID, &token_info) == CKR_OK && 
+	    !(token_info.flags & CKF_PROTECTED_AUTHENTICATION_PATH)) { 
+
+		DAEMON_ENTER ();
+		
+			if (!(token_info.flags & CKF_USER_PIN_INITIALIZED))
+				init_auth = gkr_pkcs11_auth_init_user_prompt (handle, &token_info, &new_pin, &n_new_pin);
+			/* TODO: Prompt for other 'change password' case */
+
+		DAEMON_LEAVE ();
+	}
+
+	rv = (pkcs11_lower->C_SetPIN) (handle, old_pin, n_old_pin, new_pin, n_new_pin);
+	
+	if (init_auth) {
+		DAEMON_ENTER ();
+		
+			gkr_pkcs11_auth_init_user_done (handle, &token_info, &new_pin, &n_new_pin, rv);
+			/* TODO: Done for other case */
+			
+		DAEMON_LEAVE ();
+	}
+	
+	return rv;
 }
 
 static CK_RV
@@ -509,19 +530,6 @@ auth_C_Login (CK_SESSION_HANDLE handle, CK_USER_TYPE user_type,
 	/* Try the login first, this allows NULL logins to be tried */
 	rv = (pkcs11_lower->C_Login) (handle, user_type, pin, pin_len);
 	
-	if (rv == CKR_USER_PIN_NOT_INITIALIZED) {
-
-		/* 
-		 * Try and initialize the token login, gnome-keyring modules allow 
-		 * C_SetPIN to be called on with a NULL if no user pin has yet been set.
-		 */
-		
-		if (user_type == CKU_USER) 
-			return perform_set_user_pin (handle, NULL, 0, NULL, 0, TRUE);
-		
-		return rv;
-	}
-
 	/* See if we can help the login to work */
 	if (rv != CKR_PIN_INCORRECT)
 		return rv;
