@@ -27,21 +27,46 @@
 
 #include <string.h>
 
-static GThreadPool *thread_pool = NULL;
-static GAsyncQueue *completed_queue = NULL;
-static guint completed_id = 0;
+typedef struct _GP11CallSource GP11CallSource;
 
-static void _gp11_call_implement_async_result (GAsyncResultIface *iface);
+static gpointer _gp11_call_parent_class = NULL;
 
-G_DEFINE_TYPE_EXTENDED (GP11Call, _gp11_call, G_TYPE_OBJECT, 0,
-        G_IMPLEMENT_INTERFACE (G_TYPE_ASYNC_RESULT, _gp11_call_implement_async_result));
+struct _GP11Call {
+	GObject parent;
+	GP11Module *module;
+	
+	/* For making the call */
+	GP11PerformFunc perform;
+	GP11CompleteFunc complete;
+	GP11Arguments *args;
+	GCancellable *cancellable;
+	GDestroyNotify destroy;
+	CK_RV rv;
+	
+	/* For result callback only */
+	gpointer object;
+	GAsyncReadyCallback callback;
+	gpointer user_data;
+};
+
+struct _GP11CallClass {
+	GObjectClass parent;
+	GThreadPool *thread_pool;
+	GAsyncQueue *completed_queue;
+	guint completed_id;
+};
+
+struct _GP11CallSource {
+	GSource source;
+	GP11CallClass *klass;
+};
 
 /* ----------------------------------------------------------------------------
  * HELPER FUNCTIONS
  */
 
 static CK_RV
-perform_call (GP11CallFunc func, GCancellable *cancellable, GP11Arguments *args)
+perform_call (GP11PerformFunc func, GCancellable *cancellable, GP11Arguments *args)
 {
 	CK_RV rv;
 	
@@ -69,69 +94,61 @@ perform_call (GP11CallFunc func, GCancellable *cancellable, GP11Arguments *args)
 	return rv;
 }
 
+static gboolean
+complete_call (GP11CompleteFunc func, GP11Arguments *args, CK_RV result)
+{
+	/* Double check a few things */
+	g_assert (args);
+	
+	/* If no complete function, then just ignore */
+	if (!func)
+		return TRUE;
+
+	return (func) (args, result);
+}
+
+
 static void
-process_async_call (gpointer data, gpointer unused)
+process_async_call (gpointer data, GP11CallClass *klass)
 {
 	GP11Call *call = GP11_CALL (data);
-	CK_ULONG pin_len;
 	
 	g_assert (GP11_IS_CALL (call));
 	
-	/* Try to login to the token, with the provided password */
-	if (call->do_login) {
-		call->do_login = FALSE;
-
-		pin_len = call->password ? strlen (call->password) : 0;
-		call->rv = (call->args->pkcs11->C_Login) (call->args->handle, CKU_USER, 
-		                                          (CK_UTF8CHAR_PTR)call->password, 
-		                                          pin_len);
-		
-		/* Fix the result so that we'll try the login again */
-		if (call->rv == CKR_PIN_INCORRECT)
-			call->rv = CKR_USER_NOT_LOGGED_IN;
-		
-	/* An actual call */
-	} else {
-		call->rv = perform_call (call->func, call->cancellable, call->args);
-	}
+	call->rv = perform_call (call->perform, call->cancellable, call->args);
 	
-	g_async_queue_push (completed_queue, call);
+	g_async_queue_push (klass->completed_queue, call);
 	
 	/* Wakeup main thread if on a separate thread */
 	g_main_context_wakeup (NULL);
 }
+
 static void 
 process_result (GP11Call *call, gpointer unused)
 {
+	gboolean stop = FALSE;
+	
 	/* Double check a few things */
 	g_assert (GP11_IS_CALL (call));
 	
 	if (call->cancellable) {
 		/* Don't call the callback when cancelled */
-		if (g_cancellable_is_cancelled (call->cancellable))
+		if (g_cancellable_is_cancelled (call->cancellable)) {
 			call->rv = CKR_FUNCTION_CANCELED;
+			stop = TRUE;
+		}
 	}
 	
 	/* 
-	 * Now if this is a session call, and the slot wants does 
-	 * auto-login, then we try to get a password and do auto login.
+	 * Hmmm, does the function want to actually be done?
+	 * If not, then queue this call again. 
 	 */
-	if (call->rv == CKR_USER_NOT_LOGGED_IN && GP11_IS_SESSION (call->object)) {
-		g_free (call->password);
-		call->password = NULL;
-		call->do_login = _gp11_slot_token_authentication (GP11_SESSION (call->object)->slot, 
-		                                                  &call->password);
-	}
-	
-	/* If we're supposed to do a login, then queue this call again */
-	if (call->do_login) {
+	if (!stop && !complete_call (call->complete, call->args, call->rv)) {
 		g_object_ref (call);
-		g_thread_pool_push (thread_pool, call, NULL);
-		return;
-	}
+		g_thread_pool_push (GP11_CALL_GET_CLASS (call)->thread_pool, call, NULL);
 	
 	/* All done, finish processing */
-	if (call->callback) {
+	} else if (call->callback) {
 		g_assert (G_IS_OBJECT (call->object));
 		(call->callback) (G_OBJECT (call->object), G_ASYNC_RESULT (call), 
 				  call->user_data);
@@ -139,41 +156,52 @@ process_result (GP11Call *call, gpointer unused)
 }
 
 static gboolean
-completed_prepare(GSource* source, gint *timeout)
+process_completed (GP11CallClass *klass)
 {
+	gpointer call;
+
+	g_assert (klass->completed_queue);
+
+	call = g_async_queue_try_pop (klass->completed_queue);
+	if (call) {
+		process_result (call, NULL);
+		g_object_unref (call);
+		return TRUE;
+	}
+	
+	return FALSE;
+}
+
+static gboolean
+completed_prepare(GSource* base, gint *timeout)
+{
+	GP11CallSource *source = (GP11CallSource*)base;
 	gboolean have;
-	g_assert (completed_queue);
-	have = g_async_queue_length (completed_queue) > 0;
+	
+	g_assert (source->klass->completed_queue);
+	have = g_async_queue_length (source->klass->completed_queue) > 0;
 	*timeout = have ? 0 : -1;
 	return have;
 }
 
 static gboolean
-completed_check(GSource* source)
+completed_check(GSource* base)
 {
-	g_assert (completed_queue);
-	return g_async_queue_length (completed_queue) > 0;
+	GP11CallSource *source = (GP11CallSource*)base;
+	g_assert (source->klass->completed_queue);
+	return g_async_queue_length (source->klass->completed_queue) > 0;
 }
 
 static gboolean
-completed_dispatch(GSource* source, GSourceFunc callback, gpointer user_data)
+completed_dispatch(GSource* base, GSourceFunc callback, gpointer user_data)
 {
-	gpointer *call;
-	
-	g_assert (completed_queue);
-	g_assert (callback);
-	
-	call = g_async_queue_try_pop (completed_queue);
-	if (call) {
-		((GFunc)callback) (call, user_data);
-		g_object_unref (call);
-	}
-
+	GP11CallSource *source = (GP11CallSource*)base;
+	process_completed (source->klass);
 	return TRUE;
 }
 
 static void
-completed_finalize(GSource* source)
+completed_finalize(GSource* base)
 {
 	
 }
@@ -199,6 +227,10 @@ static void
 _gp11_call_finalize (GObject *obj)
 {
 	GP11Call *call = GP11_CALL (obj);
+	
+	if (call->module)
+		g_object_unref (call->module);
+	call->module = NULL;
 
 	if (call->object)
 		g_object_unref (call->object);
@@ -213,10 +245,6 @@ _gp11_call_finalize (GObject *obj)
 	call->destroy = NULL;
 	call->args = NULL;
 	
-	if (call->password)
-		g_free (call->password);
-	call->password = NULL;
-	
 	G_OBJECT_CLASS (_gp11_call_parent_class)->finalize (obj);
 }
 
@@ -245,34 +273,98 @@ static void
 _gp11_call_class_init (GP11CallClass *klass)
 {
 	GObjectClass *gobject_class = (GObjectClass*)klass;
-	GMainContext *context;
-	GError *err = NULL;
-	GSource *src;
 
 	_gp11_call_parent_class = g_type_class_peek_parent (klass);
 	gobject_class->finalize = _gp11_call_finalize;
-	
-	g_assert (!thread_pool);
-	thread_pool = g_thread_pool_new ((GFunc)process_async_call, NULL, -1, FALSE, &err);
-	if (!thread_pool) {
+}
+
+static void
+_gp11_call_base_init (GP11CallClass *klass)
+{
+	GP11CallSource *source;
+	GMainContext *context;
+	GError *err = NULL;
+
+	klass->thread_pool = g_thread_pool_new ((GFunc)process_async_call, klass, 16, FALSE, &err);
+	if (!klass->thread_pool) {
 		g_critical ("couldn't create thread pool: %s", 
 		            err && err->message ? err->message : "");
 		return;
 	}
 	
-	g_assert (!completed_queue);
-	completed_queue = g_async_queue_new_full (g_object_unref);
-	g_assert (completed_queue);
+	klass->completed_queue = g_async_queue_new_full (g_object_unref);
+	g_assert (klass->completed_queue);
 	
 	context = g_main_context_default ();
 	g_assert (context);
 	
 	/* Add our idle handler which processes other tasks */
-	g_assert (!completed_id);
-	src = g_source_new (&completed_functions, sizeof (GSource));
-	completed_id = g_source_attach (src, context);
-	g_source_set_callback (src, (GSourceFunc)process_result, NULL, NULL);
-	g_source_unref (src);
+	source = (GP11CallSource*)g_source_new (&completed_functions, sizeof (GP11CallSource));
+	source->klass = klass;
+	klass->completed_id = g_source_attach ((GSource*)source, context);
+	g_source_set_callback ((GSource*)source, NULL, NULL, NULL);
+	g_source_unref ((GSource*)source);
+}
+
+static void
+_gp11_call_base_finalize (GP11CallClass *klass)
+{
+	GMainContext *context;
+	GSource *src;
+
+	if (klass->thread_pool) {
+		g_assert (g_thread_pool_unprocessed (klass->thread_pool) == 0);
+		g_thread_pool_free (klass->thread_pool, FALSE, TRUE);
+		klass->thread_pool = NULL;
+	}
+	
+	if (klass->completed_id) {
+		context = g_main_context_default ();
+		g_return_if_fail (context);
+		
+		src = g_main_context_find_source_by_id (context, klass->completed_id);
+		g_assert (src);
+		g_source_destroy (src);
+		klass->completed_id = 0;
+	}
+	
+	if (klass->completed_queue) {
+		g_assert (g_async_queue_length (klass->completed_queue));
+		g_async_queue_unref (klass->completed_queue);
+		klass->completed_queue = NULL;
+	}
+}
+
+GType
+_gp11_call_get_type (void) 
+{
+	static volatile gsize type_id__volatile = 0;
+	
+	if (g_once_init_enter (&type_id__volatile)) {
+		
+		static const GTypeInfo type_info = {
+			sizeof (GP11CallClass), 
+			(GBaseInitFunc)_gp11_call_base_init, 
+			(GBaseFinalizeFunc)_gp11_call_base_finalize, 
+			(GClassInitFunc)_gp11_call_class_init, 
+			(GClassFinalizeFunc)NULL, 
+			NULL,   // class_data 
+			sizeof (GP11Call), 
+			0,      // n_preallocs 
+			(GInstanceInitFunc)_gp11_call_init, 
+		};
+
+		static const GInterfaceInfo interface_info = {
+			(GInterfaceInitFunc)_gp11_call_implement_async_result
+		};
+
+		GType type_id = g_type_register_static (G_TYPE_OBJECT, "_GP11Call", &type_info, 0); 
+		g_type_add_interface_static (type_id, G_TYPE_ASYNC_RESULT, &interface_info);
+
+		g_once_init_leave (&type_id__volatile, type_id);
+	}
+	
+	return type_id__volatile;
 }
 
 /* ----------------------------------------------------------------------------
@@ -282,72 +374,36 @@ _gp11_call_class_init (GP11CallClass *klass)
 void
 _gp11_call_uninitialize (void)
 {
-	GMainContext *context;
-	GSource *src;
 
-	if (thread_pool) {
-		g_thread_pool_free (thread_pool, FALSE, TRUE);
-		thread_pool = NULL;
-	}
-	
-	if (completed_id) {
-		context = g_main_context_default ();
-		g_return_if_fail (context);
-		
-		src = g_main_context_find_source_by_id (context, completed_id);
-		g_assert (src);
-		g_source_destroy (src);
-		completed_id = 0;
-	}
-	if (completed_queue) {
-		g_async_queue_unref (completed_queue);
-		completed_queue = NULL;
-	}
 }
 
 gboolean
-_gp11_call_sync (gpointer object, gpointer func, gpointer data, 
-                 GCancellable *cancellable, GError **err)
+_gp11_call_sync (gpointer object, gpointer perform, gpointer complete, 
+                 gpointer data, GCancellable *cancellable, GError **err)
 {
 	GP11Arguments *args = (GP11Arguments*)data;
-	gchar *password = NULL;
 	GP11Module *module = NULL;
-	CK_ULONG pin_len;
 	CK_RV rv;
 	
 	g_assert (G_IS_OBJECT (object));
-	g_assert (func);
+	g_assert (perform);
 	g_assert (args);
 	
 	g_object_get (object, "module", &module, "handle", &args->handle, NULL);
 	g_assert (GP11_IS_MODULE (module));
-	
-	args->pkcs11 = module->funcs;
-	g_object_unref (module);
-	
-	rv = perform_call ((GP11CallFunc)func, cancellable, args);
-		
-	/* 
-	 * Now if this is a session call, and the slot wants does 
-	 * auto-login, then we try to get a password and do auto login.
-	 */
-	if (rv == CKR_USER_NOT_LOGGED_IN && GP11_IS_SESSION (object)) {
-		
-		do {
-			if (!_gp11_slot_token_authentication (GP11_SESSION (object)->slot, 
-			                                      &password)) {
-				rv = CKR_USER_NOT_LOGGED_IN;
-			} else {
-				pin_len = password ? strlen (password) : 0; 
-				rv = (args->pkcs11->C_Login) (args->handle, CKU_USER, 
-				                              (CK_UTF8CHAR_PTR)password, pin_len);
-			}
-		} while (rv == CKR_PIN_INCORRECT);
 
-		/* If we logged in successfully then try again */
-		if (rv == CKR_OK)
-			rv = perform_call ((GP11CallFunc)func, cancellable, args);
-	}
+	/* We now hold a reference to module until below */
+	args->pkcs11 = gp11_module_get_functions (module);
+	g_assert (args->pkcs11);
+	
+	do {
+		rv = perform_call (perform, cancellable, args);
+		if (rv == CKR_FUNCTION_CANCELED)
+			break;
+		
+	} while (!complete_call (complete, args, rv));
+		
+	g_object_unref (module);
 
 	if (rv == CKR_OK)
 		return TRUE;
@@ -357,15 +413,14 @@ _gp11_call_sync (gpointer object, gpointer func, gpointer data,
 }
 
 gpointer
-_gp11_call_async_prep (gpointer object, gpointer cb_object, gpointer func, 
-                       gsize args_size, gpointer destroy)
+_gp11_call_async_prep (gpointer object, gpointer cb_object, gpointer perform, 
+                       gpointer complete, gsize args_size, gpointer destroy)
 {
 	GP11Arguments *args;
-	GP11Module *module;
 	GP11Call *call;
 
-	g_assert (G_IS_OBJECT (object));
-	g_assert (func);
+	g_assert (!object || G_IS_OBJECT (object));
+	g_assert (perform);
 	
 	if (!destroy)
 		destroy = g_free;
@@ -375,14 +430,10 @@ _gp11_call_async_prep (gpointer object, gpointer cb_object, gpointer func,
 	g_assert (args_size >= sizeof (GP11Arguments));
 	
 	args = g_malloc0 (args_size);
-	g_object_get (object, "module", &module, "handle", &args->handle, NULL);
-	g_assert (GP11_IS_MODULE (module));
-	args->pkcs11 = module->funcs;
-	g_object_unref (module);
-	
 	call = g_object_new (GP11_TYPE_CALL, NULL);
 	call->destroy = (GDestroyNotify)destroy;
-	call->func = (GP11CallFunc)func;
+	call->perform = (GP11PerformFunc)perform;
+	call->complete = (GP11CompleteFunc)complete;
 	call->object = cb_object;
 	g_object_ref (cb_object);
 
@@ -390,31 +441,35 @@ _gp11_call_async_prep (gpointer object, gpointer cb_object, gpointer func,
 	call->args = args;
 	call->args->call = call;
 
+	/* Setup call object if available */
+	if (object != NULL)
+		_gp11_call_async_object (call, object);
+
 	return args;
 }
 
-void 
-_gp11_call_async_short (gpointer data, GAsyncReadyCallback callback,
-                        gpointer user_data)
+void
+_gp11_call_async_object (GP11Call *call, gpointer object)
 {
-	GP11Arguments *args = (GP11Arguments*)data;
+	g_assert (GP11_IS_CALL (call));
+	g_assert (call->args);
 	
-	g_assert (GP11_IS_CALL (args->call));
+	if (call->module)
+		g_object_unref (call->module);
+	call->module = NULL;
 	
-	args->call->callback = callback;
-	args->call->user_data = user_data;
+	g_object_get (object, "module", &call->module, "handle", &call->args->handle, NULL);
+	g_assert (GP11_IS_MODULE (call->module));
+	call->args->pkcs11 = gp11_module_get_functions (call->module);
 	
-	/* Already complete, so just push it for processing in main loop */
-	g_assert (completed_queue);
-	g_async_queue_push (completed_queue, args->call);
+	/* We now hold a reference on module until finalize */
 }
 
-void
-_gp11_call_async_go (gpointer data, GCancellable *cancellable, 
-                     GAsyncReadyCallback callback, gpointer user_data)
+GP11Call*
+_gp11_call_async_ready (gpointer data, GCancellable *cancellable, 
+                        GAsyncReadyCallback callback, gpointer user_data)
 {
 	GP11Arguments *args = (GP11Arguments*)data;
-	
 	g_assert (GP11_IS_CALL (args->call));
 	
 	args->call->cancellable = cancellable;
@@ -426,12 +481,32 @@ _gp11_call_async_go (gpointer data, GCancellable *cancellable,
 	args->call->callback = callback;
 	args->call->user_data = user_data;
 	
-	g_assert (thread_pool);
-	g_thread_pool_push (thread_pool, args->call, NULL);
+	return args->call;
+}
+
+void
+_gp11_call_async_go (GP11Call *call)
+{
+	g_assert (GP11_IS_CALL (call));
+	g_assert (call->args->pkcs11);
+
+	/* To keep things balanced, process at one completed event */
+	process_completed(GP11_CALL_GET_CLASS (call));
+
+	g_assert (GP11_CALL_GET_CLASS (call)->thread_pool);
+	g_thread_pool_push (GP11_CALL_GET_CLASS (call)->thread_pool, call, NULL);
+}
+
+void
+_gp11_call_async_ready_go (gpointer data, GCancellable *cancellable, 
+                           GAsyncReadyCallback callback, gpointer user_data)
+{
+	GP11Call *call = _gp11_call_async_ready (data, cancellable, callback, user_data);
+	_gp11_call_async_go (call);
 }
 
 gboolean
-_gp11_call_basic_finish (gpointer object, GAsyncResult *result, GError **err)
+_gp11_call_basic_finish (GAsyncResult *result, GError **err)
 {
 	CK_RV rv;
 	
@@ -444,3 +519,22 @@ _gp11_call_basic_finish (gpointer object, GAsyncResult *result, GError **err)
 	g_set_error (err, GP11_ERROR, rv, "%s", gp11_message_from_rv (rv));
 	return FALSE;	
 }
+
+void
+_gp11_call_async_short (GP11Call *call, CK_RV rv)
+{
+	g_assert (GP11_IS_CALL (call));
+	
+	call->rv = rv;
+
+	/* Already complete, so just push it for processing in main loop */
+	g_assert (GP11_CALL_GET_CLASS (call)->completed_queue);
+	g_async_queue_push (GP11_CALL_GET_CLASS (call)->completed_queue, call);
+}
+
+gpointer
+_gp11_call_get_arguments (GP11Call *call)
+{
+	g_assert (GP11_IS_CALL (call));
+	return call->args;
+}
