@@ -65,7 +65,6 @@ typedef struct _GP11ModulePrivate {
 	gboolean finalized;
 	GHashTable *open_sessions;
 	gboolean auto_authenticate;
-	GP11TokenInfo *token_info;
 } GP11ModulePrivate;
 
 #define GP11_MODULE_GET_DATA(o) \
@@ -86,24 +85,6 @@ typedef struct _SessionPool {
  * HELPERS
  */
 
-static guint
-ulong_hash (gconstpointer v)
-{
-	const signed char *p = v;
-	guint32 i, h = *p;
-
-	for(i = 0; i < sizeof (gulong); ++i)
-		h = (h << 5) - h + *(p++);
-
-	return h;
-}
-
-static gboolean
-ulong_equal (gconstpointer v1, gconstpointer v2)
-{
-	return *((const gulong*)v1) == *((const gulong*)v2);
-}
-
 static CK_RV
 create_mutex (void **mutex)
 {
@@ -302,7 +283,7 @@ static void
 create_session_table (GP11ModulePrivate *pv)
 {
 	if (!pv->open_sessions)
-		pv->open_sessions = g_hash_table_new_full (ulong_hash, ulong_equal, g_free, free_session_pool);
+		pv->open_sessions = g_hash_table_new_full (_gp11_ulong_hash, _gp11_ulong_equal, g_free, free_session_pool);
 }
 
 CK_SESSION_HANDLE
@@ -702,6 +683,55 @@ gp11_module_new (CK_FUNCTION_LIST_PTR funcs)
 	return g_object_new (GP11_TYPE_MODULE, "functions", funcs, NULL);
 }
 
+/**
+ * gp11_module_equal:
+ * @module1: A pointer to the first GP11Module
+ * @module2: A pointer to the second GP11Module
+ * 
+ * Checks equality of two modules. Two GP11Module objects can point to the same 
+ * underlying PKCS#11 module.
+ * 
+ * Return value: TRUE if module1 and module2 are equal. FALSE if either is not a GP11Module.
+ **/
+gboolean
+gp11_module_equal (gconstpointer module1, gconstpointer module2)
+{
+	GP11ModuleData *data1, *data2;
+
+	if (module1 == module2)
+		return TRUE;
+	if (!GP11_IS_MODULE (module1) || !GP11_IS_MODULE (module2))
+		return FALSE;
+	
+	data1 = GP11_MODULE_GET_DATA (module1);
+	data2 = GP11_MODULE_GET_DATA (module2);
+	
+	return data1->funcs == data2->funcs;
+}
+
+/**
+ * gp11_module_hash:
+ * @module: A pointer to a GP11Module
+ * 
+ * Create a hash value for the GP11Module. 
+ * 
+ * This function is intended for easily hashing a GP11Module to add to 
+ * a GHashTable or similar data structure.
+ * 
+ * Return value: An integer that can be used as a hash value, or 0 if invalid.
+ **/
+guint
+gp11_module_hash (gconstpointer module)
+{
+	GP11ModuleData *data;
+	
+	g_return_val_if_fail (GP11_IS_MODULE (module), 0);
+
+	data = GP11_MODULE_GET_DATA (module);
+	
+	return g_direct_hash (data->funcs);
+}
+
 /**
  * gp11_module_get_info:
  * @self: The module to get info for.
@@ -969,6 +999,9 @@ gp11_module_set_auto_authenticate (GP11Module *self, gboolean auto_login)
  * You can access the session in which the object was found, by using the 
  * gp11_object_get_session() function on the resulting objects.
  * 
+ * This function skips tokens that are not initialize, and makes a best effort to 
+ * find objects on valid tokens. 
+ * 
  * The function can return FALSE to stop the enumeration.
  * 
  * Return value: If FALSE then an error prevented all matching objects from being enumerated.
@@ -1040,9 +1073,19 @@ gp11_module_enumerate_objects_full (GP11Module *self, GP11Attributes *attrs,
 	slots = gp11_module_get_slots (self, TRUE);
 	
 	for (l = slots; ret && !stop && l; l = g_list_next (l)) {
-		session = gp11_slot_open_session (l->data, CKF_SERIAL_SESSION, err);
+		session = gp11_slot_open_session (l->data, CKF_SERIAL_SESSION, &error);
 		if (!session) {
-			ret = FALSE;
+			g_return_val_if_fail (error != NULL, FALSE);
+			
+			/* Ignore these errors when enumerating */
+			if (error->code == CKR_USER_PIN_NOT_INITIALIZED) {
+				g_clear_error (&error);
+				
+			} else {
+				ret = FALSE;
+				g_propagate_error (err, error);
+				error = NULL;
+			}
 			continue;
 		}
 		
